<?php

declare(strict_types=1);

namespace Drupal\menu_ui_async_widget\Hook;

use Drupal\Component\Utility\Html;
use Drupal\Core\DependencyInjection\ClassResolverInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityFormInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Hook\Attribute\Hook;
use Drupal\Core\Hook\Attribute\RemoveHook;
use Drupal\Core\Menu\MenuParentFormSelectorInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\menu_ui\Hook\MenuUiHooks;
use Drupal\node\NodeInterface;
use Drupal\node\NodeTypeInterface;
use Drupal\system\MenuInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

class Hooks implements ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * Class constructor.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected AccountProxyInterface $currentUser,
    protected MenuParentFormSelectorInterface $menuParentFormSelector,
    protected EntityRepositoryInterface $entityRepository,
    protected EntityFieldManagerInterface $entityFieldManager,
    protected ModuleHandlerInterface $moduleHandler,
    protected ClassResolverInterface $classResolver
  ) {}

  /**
   * {@inheritDoc}
   *
   * @noinspection PhpParamsInspection
   */
  public static function create(ContainerInterface $container) : static {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('current_user'),
      $container->get('menu.parent_form_selector'),
      $container->get('entity.repository'),
      $container->get('entity_field.manager'),
      $container->get('module_handler'),
      $container->get('class_resolver')
    );
  }

  /**
   * Implements RemoveHook.
   *
   * Ensure {@see MenuUiHooks::formNodeFormAlter()} is never run.
   */
  #[RemoveHook('form_node_form_alter', class: MenuUiHooks::class, method: 'formNodeFormAlter')]

  /**
   * Implements hook_help().
   */
  #[Hook('help')]
  public function help($route_name, RouteMatchInterface $route_match): ?string {
    if ($route_name == 'help.page.menu_ui_async_widget') {
      $output = '<h2>' . $this->t('About') . '</h2>';
      $output .= '<p>' . $this->t('This module removes the menu selection UI from node forms and replaces it with a button that loads the full UI if the user wants to add or edit the menu link associated with the node.');
      $output .= '<p>' . $this->t('In the scenario this module was developed for, two available menus with about 1500 links in each, the performance increase of loading the node form was around 650%.') . '</p>';
      $output .= '<p>' . $this->t('The impact of using this module will vary depending on your specific setup. If you don\'t have large menus you will probably not see any performance gains. Instead you will force the user to go through an extra step to add/edit the menu link.') . '</p>';
      return $output;
    }
    return NULL;
  }

  /**
   * Implements hook_form_BASE_FORM_ID_alter().
   *
   * Replace the default widget with the new async widget.
   *
   * @see MenuUiHooks::formNodeFormAlter()
   */
  #[Hook('form_node_form_alter')]
  public function formNodeFormAlter(&$form, FormStateInterface $form_state) : void {
    /** @var EntityFormInterface $form_object */
    $form_object = $form_state->getFormObject();
    /** @var NodeInterface $node */
    $node = $form_object->getEntity();
    $defaults = menu_ui_get_menu_link_defaults($node);
    /** @var NodeTypeInterface $node_type */
    $node_type = $node->get('type')->entity;

    // Do not alter the form if 'menu_ui_async_widget' has not been configured
    // for the node type, instead we invoke the original form alter from
    // the 'menu_ui' module.
    $use_async_widget = $node_type->getThirdPartySetting('menu_ui_async_widget', 'use_async_widget', FALSE);
    if (!$use_async_widget) {
      $this->classResolver->getInstanceFromDefinition('Drupal\menu_ui\Hook\MenuUiHooks')->formNodeFormAlter($form, $form_state);
      return;
    }

    $form['menu'] = [
      '#type' => 'details',
      '#title' => t('Menu settings'),
      '#access' => $this->currentUser->hasPermission('administer menu'),
      '#open' => (bool) $defaults['id'],
      '#group' => 'advanced',
      '#attached' => [
        'library' => ['menu_ui/drupal.menu_ui'],
      ],
      '#tree' => TRUE,
      '#weight' => -2,
      '#attributes' => ['class' => ['menu-link-form']],
    ];

    // Has the user clicked to customize the menu link?
    if ($form_state->get('menu_ui_async_widget')) {
      // Generate a list of possible parents (not including this link or descendants).
      $type_menus_ids = $node_type->getThirdPartySetting('menu_ui', 'available_menus');
      if (empty($type_menus_ids)) {
        return;
      }
      /** @var MenuInterface[] $type_menus */
      $type_menus = $this->entityTypeManager->getStorage('menu')->loadMultiple($type_menus_ids);
      $available_menus = [];
      foreach ($type_menus as $menu) {
        $available_menus[$menu->id()] = $menu->label();
      }
      if ($defaults['id']) {
        $default = $defaults['menu_name'] . ':' . $defaults['parent'];
      }
      else {
        $default = $node_type->getThirdPartySetting('menu_ui', 'parent', 'main:');
      }
      $parent_element = $this->menuParentFormSelector->parentSelectElement($default, $defaults['id'], $available_menus);
      // If no possible parent menu items were found, there is nothing to display.
      if (empty($parent_element)) {
        return;
      }

      $form['menu']['container'] = [
        '#type' => 'container',
        '#parents' => ['menu'],
        '#tree' => TRUE,
      ];
      $form['menu']['container']['enabled'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Provide a menu link'),
        '#default_value' => (int) (bool) $defaults['id'],
      ];

      $form['menu']['container']['link'] = [
        '#type' => 'container',
        '#parents' => ['menu'],
        '#states' => [
          'invisible' => [
            'input[name="menu[enabled]"]' => ['checked' => FALSE],
          ],
        ],
      ];

      // Populate the element with the link data.
      foreach (['id', 'entity_id'] as $key) {
        $form['menu']['container']['link'][$key] = ['#type' => 'value', '#value' => $defaults[$key]];
      }

      $form['menu']['container']['link']['title'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Menu link title'),
        '#default_value' => $defaults['title'],
        '#maxlength' => $defaults['title_max_length'],
      ];

      $form['menu']['container']['link']['description'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Description'),
        '#default_value' => $defaults['description'],
        '#description' => t('Shown when hovering over the menu link.'),
        '#maxlength' => $defaults['description_max_length'],
      ];

      $form['menu']['container']['link']['menu_parent'] = $parent_element;
      $form['menu']['container']['link']['menu_parent']['#title'] = $this->t('Parent link');
      $form['menu']['container']['link']['menu_parent']['#attributes']['class'][] = 'menu-parent-select';

      $form['menu']['container']['link']['weight'] = [
        '#type' => 'number',
        '#title' => $this->t('Weight'),
        '#default_value' => $defaults['weight'],
        '#description' => t('Menu links with lower weights are displayed before links with higher weights.'),
      ];
    }
    else {
      $wrapper_id = Html::getUniqueId('menu-ui-async-widget-wrapper');
      $form['menu']['container'] = [
        '#type' => 'container',
      ];
      $form['menu']['container']['submit'] = [
        '#type' => 'submit',
        '#value' => ((bool) $defaults['id']) ? $this->t('Edit menu link') : $this->t('Add menu link'),
        '#limit_validation_errors' => [],
        '#submit' => [[$this, 'nodeFormSubmit']],
        '#ajax' => [
          'callback' => [$this, 'ajaxCallback'],
          'wrapper' => $wrapper_id,
        ],
        '#prefix' => "<span id=\"$wrapper_id\">",
        '#suffix' => "</span>",
      ];
      if ($defaults['title']) {
        // We need these fields here to make the drupal.menu_ui library display
        // the correct summary before we load the complete form.
        $form['menu']['container']['enabled'] = [
          '#type' => 'checkbox',
          '#value' => (int) (bool) $defaults['id'],
          '#prefix' => '<span class="js-form-item-menu-enabled hidden">',
          '#suffix' => '</span>',
        ];
        $form['menu']['container']['title'] = [
          '#type' => 'textfield',
          '#value' => $defaults['title'],
          '#prefix' => '<span class="js-form-item-menu-title hidden">',
          '#suffix' => '</span>',
        ];
      }
    }

    foreach (array_keys($form['actions']) as $action) {
      if ($action != 'preview' && isset($form['actions'][$action]['#type']) && $form['actions'][$action]['#type'] === 'submit') {
        $form['actions'][$action]['#submit'][] = 'menu_ui_form_node_form_submit';
      }
    }

    $form['#entity_builders'][] = 'menu_ui_node_builder';
  }

  /**
   * Form submission handler for add/edit menu item on the node form.
   */
  public function nodeFormSubmit(array &$form, FormStateInterface $form_state) : void {
    $form_state->set('menu_ui_async_widget', TRUE);
    $form_state->setRebuild();
  }

  /**
   * AJAX form callback for loading additional form elements.
   */
  public static function ajaxCallback(array $form, FormStateInterface $form_state) : array {
    return $form['menu']['container'];
  }

  /**
   * Implements hook_form_FORM_ID_alter().
   *
   * Adds menu options to the node type form.
   *
   * @see \Drupal\node\Form\NodeTypeForm::form()
   * @see menu_ui_form_node_type_form_builder()
   */
  #[Hook('form_node_type_form_alter')]
  public function nodeTypeFormAlter(&$form, FormStateInterface $form_state, $form_id) : void {
    /** @var EntityFormInterface $form_object */
    $form_object = $form_state->getFormObject();
    /** @var NodeTypeInterface $type */
    $type = $form_object->getEntity();

    $form['menu']['use_async_widget'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use asynchronous loading of menu widget'),
      '#default_value' => $type->getThirdPartySetting('menu_ui_async_widget', 'use_async_widget', FALSE),
      '#description' => $this->t('If checked, an asynchronous widget will be used to add/edit a menu link on node forms.'),
    ];

    $form['#entity_builders'][] = [$this, 'nodeTypeFormBuilder'];
  }

  /**
   * Entity builder for the node type form with menu options.
   */
  public static function nodeTypeFormBuilder($entity_type, NodeTypeInterface $type, &$form, FormStateInterface $form_state) : void {
    $type->setThirdPartySetting('menu_ui_async_widget', 'use_async_widget', $form_state->getValue('use_async_widget'));
  }

}
